Unit DebugMe;

(*  ***  CRUDE but VERY useful debugging message system  ****
                 For Dos Target applications.

                     by Brian Dockerty
                    Dockerty Electronics
                      bdock@wave.co.nz

                REVISION 1.1 - See notes below.

    ***********************************************************

    Provides a SIMPLE Error Handler and debugging message stack.
    This stack shows the function / procedure nesting level at the
    time of the error.

    Debug message dump & optional LOG FILE dump are provided to assist with
    the error post-mortem.

    ***********************************************************

    YES, this DOES slow does execution, BUT it has also helped me
    locate some VERY frustrating bugs with ease.

    If you do not ENABLE the message system then the overhead is
    significantly reduced, so you shouldn't have any problems
    leaving the debugging code in ALL of your source.

    The GREATEST gain from using this is the ERROR LOG FILE.
    It helps you locate those nasty little bugs that only seem
    to occur every few hours.

    During the debugging phase you can ship a copy of you software
    to a customer with the debugging ENABLED and ask them to fax or
    email the "Errors.Txt" file when they have a problem.

    ***********************************************************

    TO USE :

    Install the Error Handler at the start of your program.

    Uninstall the Error Handler at the end of your program.

    To Enable the message system Set DEBUG_MESSAGES_ENABLED to TRUE

    To Enable error logging to "Errors.Txt" text file, SET ERROR_LOG to TRUE

    In EACH Procedure or Function (or just those you want to watch)

      INSERT      Set_Debug_Message(your procedure name)
                           as the FIRST statement
                                  and
      INSERT      Remove_Debug_Message
                           as the LAST statement

    Upon error a list of the procedure / function nesting will be displayed
    and, if logging enabled, appended to "Errors.Txt" text file.

    ***********************************************************
                       WARNINGS / PRECAUTIONS

        To successfully use this unit to locate errors,
        you must NOT use EXIT to leave a procedure or function that is
        being tracked as this will create a message level leak by
        leaving without having executed the Remove_Debug_Message that
        is at the end of the prodecure / function.

        The message system is INTERRUPT SAFE as it PRESERVES the interrupt
        flag AND masks interrupts DURING message level updates.

        IF you install ANY custom HARDWARE interrupt handlers,
        you will need to modify the Error_Handler procedure to
        uninstall them as required.

    ***********************************************************

                 REVISION 1.1 - 20 July 2000
    Fix CRITICAL FAULT. - Switched to using DPMI GETINTERRUPTSTATE
    to get interrupt state as the CPU FLAGS do NOT reflect the
    TRUE STATE of the Interrupt flag in a DPMI environment.

    I had NOT read the DPMI 0.9 specification, until recently,
    and I was caught out by one of the most basic of protected mode issues.

    ***********************************************************

        If ANYONE has any other useful debugging tools then
                        Please contact me.

    ***********************************************************

*)

Interface

Procedure Set_Debug_Message(DebugMessage : string);     {add message to message stack}
Procedure Remove_Debug_Message;                         {remove message from message stack}
Procedure Install_Error_Handler;               {install the error handler}
Procedure UnInstall_Error_Handler;             {remove the rrror handler}

Const
  MessageHeapSize = 1024;                      {maximum message nesting level - this should be sufficent!}
  Debug_Message_Count    : Longint = 0;        {current message stack level}
  Debug_Messages_Enabled : boolean = false;    {set this to enable & clear this to freeze the message stack}
  Error_Log : boolean = false;                 {error log file enable flag}

Var
  Debug_Messages : array[1..MessageHeapSize] of string; {debug message stack structure}

Implementation

uses
  Dpmi, ErrCodes;

const
  Error_Handler_Installed : boolean = false;   {error handler installation status}

Var
  X : Longint;
  OldExitProc : Pointer;

procedure EnableInterrupts; inline($FB);       {sti}

procedure DisableInterrupts; inline($FA);      {cli}

Procedure Set_Debug_Message(DebugMessage : string);     {add new message to message stack}
Var
  Enabled : boolean;
begin
  If Debug_Messages_Enabled then
  begin
    Enabled := GetInterruptState;           {get current interrupt state to preserve it}
    DisableInterrupts;
    If Debug_Message_Count < MessageHeapSize then  {ensure there is room first}
    begin
      Inc(Debug_Message_Count);
      Debug_Messages[Debug_Message_Count] := DebugMessage;  {put message on message stack}
    end;
    if Enabled then EnableInterrupts;       {if Interrupts were enabled then re-enable them}
  end;
end;

Procedure Remove_Debug_Message;             {remove current error message from message stack}
Var
  Enabled : boolean;
begin
  If Debug_Messages_Enabled then
  begin
    Enabled := GetInterruptState;           {get current interrupt state to preserve it}
    DisableInterrupts;
    if Debug_Message_Count > 0 then dec(Debug_Message_Count); {remove message from message stack}
    if Enabled then EnableInterrupts;       {if Interrupts were enabled then re-enable them}
  end;
end;

procedure Error_Handler; Forward;      {forward declaration - defined later}

Procedure Install_Error_Handler;       {install the error handler}
begin
  if not Error_Handler_Installed then
  begin
    OldExitProc := ExitProc;           {save old error handler address}
    ExitProc := @Error_Handler;
    Error_Handler_Installed := true;
  end;
end;

Procedure UnInstall_Error_Handler;     {remove the rrror handler}
begin
  if Error_Handler_Installed then      {if error handler was installed}
  begin
    ExitProc := OldExitProc;           {restore original error handler}
    Error_Handler_Installed := false;
  end;
end;

procedure Error_Handler;               {display semi-meaningful error messages}
Var
  X : Longint;
  F : text;
begin
  Debug_Messages_Enabled := false;     {freeze the message stack}
  Uninstall_Error_Handler;             {remove error handler}

  Writeln('Error ', ExitCode, ' ', Error_msg (ExitCode)); {display error message}
  Writeln('***** DEBUG MESSAGE STACK DUMP *****');
  IoResult;                            {clear any pending io error}
{$I-}                                  {disable all IO checking for log file append}
  If Error_Log then
  begin
    Assign(f,'Errors.txt');            {append error to error log}
    append(f);
    Writeln(f,'Error ', ExitCode, ' ', Error_msg (ExitCode)); {log error message}
    Writeln(f,'***** DEBUG MESSAGE STACK DUMP *****');
  end;

  If Debug_Message_Count > 0 then
  begin
    for X := 1 to Debug_Message_Count do
    begin
      Writeln(Debug_Messages[X]);      {display User Debug Task Message}
      if Error_Log then Writeln(f, Debug_Messages[X]);   {log User Error Task Message}
    end;
  end;

  if Error_Log then                    {if error log then}
  begin
    Writeln(f);                        {add a blank terminator line to file}
    Close(f);                          {and close it}
    Writeln('******  Please Print The ERROR LOG  ******');
  end;
  IoResult;                            {flush ioresult}
{$I+}
end;

end.